/********************************************************\
|*                                                      *|
|*                  VeriFinger 4.2 SDK                  *|
|*                                                      *|
|*  ScanManX.cpp                                        *|
|*    Header file for extended Scanner Manager features *|
|*                                                      *|
|*  Copyright (C) 2002-2004 Neurotechnologija           *|
|*                                                      *|
\********************************************************/

#include "System.h"
#include "ScanManX.h"
#include "Utils.h"

////////////////////////////////
//                            //
// Error and result functions //
//                            //
////////////////////////////////

string SMErrorToString(INT error)
{
	if(error >= 0) return emptyString;
	switch(error)
	{
		case SME_FAILED: return TEXT("Failed");
		case SME_OUT_OF_MEMORY: return TEXT("Out of memory");
		case SME_NOT_INITIALIZED: return TEXT("Not initialized");
		case SME_ARGUMENT_NULL: return TEXT("Argument null");
		case SME_INVALID_ARGUMENT: return TEXT("Invalid argument");
		case SME_INVALID_PARAMETER: return TEXT("Invalid parameter");
		case SME_PARAMETER_READ_ONLY: return TEXT("Parameter is read only");
		case SME_ALREADY_CAPTURING: return TEXT("Already capturing");
		case SME_NOT_CAPTURING: return TEXT("Not capturing");
		default: return IntToString(error);
	}
}

string SMResultToString(INT result)
{
	if(SMFailed(result)) return SMErrorToString(result);
	else return result == SME_OK ? STRING("OK") : Format(TEXT("OK (%d)"), result);
}

void SMCheckResult(INT result)
{
	if(SMFailed(result)) throw EXCEPTION(Format(TEXT("ScanMan error:\n\tCode = %d\n\tDescription = %s"),
		result, (LPCTSTR)SMErrorToString(result)));
}

//////////////////////////
//                      //
// Parameters functions //
//                      //
//////////////////////////

bool SMParameterOfType(INT reserved, INT parameter, INT type)
{
	return SMGetParameterType(reserved, parameter) == type;
}

void SMCheckParameterType(INT reserved, INT parameter, INT type)
{
	if(!SMParameterOfType(reserved, parameter, type)) throw exception("ScanMan parameter is of wrong type");
}

// Get

DWORD SMGetDWordParameter(INT reserved, INT parameter)
{
	SMCheckParameterType(reserved, parameter, SM_TYPE_DWORD);
	DWORD value;
	SMCheckResult(SMGetParameter(reserved, parameter, &value));
	return value;
}

string SMGetStringParameter(INT reserved, INT parameter)
{
	SMCheckParameterType(reserved, parameter, SM_TYPE_STRING);
	int l = SMGetParameter(reserved, parameter, NULL);
	SMCheckResult(l);
	char *value;
	value = new char[l + 1];
	l = SMGetParameter(reserved, parameter, value);
	string s;
	if(SMSucceeded(l)) s = AnsiToString(value);
	delete[] value;
	SMCheckResult(l);
	return s;
}

// Get general

INT SMGetParameterType(INT reserved, INT parameter)
{
	INT result = SMGetParameter(reserved, SMP_TYPE, (VOID*)parameter);
	SMCheckResult(result);
	return result;
}

string SMGetName(INT reserved)
{
	return SMGetStringParameter(reserved, SMP_NAME);
}

DWORD SMGetVersionHigh(INT reserved)
{
	return SMGetDWordParameter(reserved, SMP_VERSION_HIGH);
}

DWORD SMGetVersionLow(INT reserved)
{
	return SMGetDWordParameter(reserved, SMP_VERSION_LOW);
}

string SMGetCopyright(INT reserved)
{
	return SMGetStringParameter(reserved, SMP_COPYRIGHT);
}

///////////////////////////////////
//                               //
// Scanner enumeration functions //
//                               //
///////////////////////////////////

string SMGetScannerId(INT index)
{
	INT l = SMGetScannerId(index, NULL, 0);
	if(SMSucceeded(l))
	{
		CHAR * id = new CHAR[l];
		l = SMGetScannerId(index, id, l);
		string s;
		if(SMSucceeded(l)) s = AnsiToString(id);
		delete[] id;
		if(SMSucceeded(l)) return s;
	}
	throw exception("Can not retrieve scanner id");
}

void SMGetScannerIds(strings &ids)
{
	ids.Clear();
	INT c = SMGetScannerCount();
	if(SMSucceeded(c))
	{
		for(int i = 0; i < c; i++)
			try
			{
				ids.Add(SMGetScannerId(i));
			}
			catch(exception &)
			{
			}
	}
}

/////////////////////////
//                     //
// Capturing functions //
//                     //
/////////////////////////

INT SMStartCapturing(const string &scanner_id, SM_IMAGE_PROC * image_proc, SM_STATE_PROC * state_proc, VOID * param)
{
#ifdef _UNICODE
	int l = StringToAnsi(scanner_id, NULL, 0);
	CHAR *id = new CHAR[l];
	StringToAnsi(scanner_id, id, l);
	INT result = SMStartCapturing(id, image_proc, state_proc, param);
	delete[] id;
	return result;
#else
	return SMStartCapturing((LPCTSTR)scanner_id, image_proc, state_proc, param);
#endif
}

INT SMStopCapturing(const string &scanner_id)
{
#ifdef _UNICODE
	int l = StringToAnsi(scanner_id, NULL, 0);
	CHAR *id = new CHAR[l];
	INT result = StringToAnsi(scanner_id, id, l);
	SMStopCapturing(id);
	delete[] id;
	return result;
#else
	return SMStopCapturing((LPCTSTR)scanner_id);
#endif
}

INT SMStopCapturing()
{
	return SMStopCapturing(NULL);
}

///////////////////////////////
//                           //
// Event and state functions //
//                           //
///////////////////////////////

string SMEventToString(DWORD event, const string &scanner_id)
{
	string s;
	switch(event)
	{
		case SM_EVENT_PLUGGED: s = TEXT("Plugged"); break;
		case SM_EVENT_UNPLUGGED: s = TEXT("Unplugged"); break;
		case SM_EVENT_ERROR: s = TEXT("Error"); break;
		default: s = IntToString(event);
	}
	return scanner_id.IsEmpty() ? s : s + TEXT(": ") + scanner_id;
}

string SMStateToString(DWORD state, const string &scanner_id)
{
	switch(state)
	{
		case SM_STATE_FINGER_DOWN: return scanner_id.IsEmpty() ? TEXT("Finger down") : TEXT("Finger placed onto ") + scanner_id;
		case SM_STATE_FINGER_UP: return scanner_id.IsEmpty() ? TEXT("Finger up") : TEXT("Finger removed from ") + scanner_id;
		default: return IntToString(state) + (scanner_id.IsEmpty() ? TEXT("") : TEXT(" on ") + scanner_id);
	}
}
