package neurotechnologija.utils;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.image.ImageObserver;
import java.awt.image.MemoryImageSource;
import java.awt.image.PixelGrabber;
import java.io.File;

import javax.swing.JFileChooser;

/**
 * Utility functions used by VeriFinger Demo
 *
 * @version 1.0 02/07/2002
 */
public class Utils {

        /**
         * Display window centered on screen
         * */
        public static void centerOnScreen(Window wnd) {
                Dimension dim = wnd.getToolkit().getScreenSize();
                Rectangle bounds = wnd.getBounds();
                wnd.setLocation(
                        (dim.width - bounds.width) / 2,
                        (dim.height - bounds.height) / 2);
                wnd.setVisible(true);
                wnd.requestFocus();
        }

        /**
         * Center a window over a parent window
         * @param wnd window to center
         * @see java.awt.Window#setLocation
         * */
        public static void centerParent(Window wnd) {
                int x;
                int y;

                // Find out our parent
                Container parent = wnd.getParent();
                Point topLeft = parent.getLocationOnScreen();
                Dimension parentSize = parent.getSize();

                Dimension size = wnd.getSize();

                x = topLeft.x;
                y = topLeft.y;
                if (parentSize.width > size.width)
                        x = (int) ((parentSize.width - size.width) / 2) + topLeft.x;

                if (parentSize.height > size.height)
                        y = (int) ((parentSize.height - size.height) / 2) + topLeft.y;

                wnd.setLocation(x, y);
                wnd.requestFocus();
        }


        /**
         * Checks whether array contains <code>null</code> values or is equal to <code>null</null> itself
         * @return boolean true if <code>null</code> values were found in array or array is <code>null</code>
         * */
        public static boolean isNull(Object[] array) {
                if (array == null) return true;
                for (int i = 0; i < array.length; ++i) {
                        if (array[i] == null) return true;
                }
                return false;
        }

        /**
         * Select single file using file selection dialog
         * @param owner Component over which the dialog will be displayed
         * @param title file selection dialog title
         * @return String selected image file path
         * @see javax.swing#JFileChooser
         * */
        public static String selectFile(Component owner, String title) {
                String result = null;
                JFileChooser dlg = new JFileChooser();
                dlg.setDialogTitle(title);

                int returnVal = dlg.showOpenDialog(owner);
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                        result = dlg.getSelectedFile().getPath();
                }

                return result;
        }

        /**
         * Select single image file using file selection dialog
         * @return String selected image file path
         * @see javax.swing#JFileChooser
         * */
        public static String selectImage(Component owner, String title) {
                String result = null;
                JFileChooser dlg = new JFileChooser();
                dlg.setDialogTitle(title);
                ExampleFileFilter filter = new ExampleFileFilter();
                filter.addExtension("bmp");
                filter.addExtension("gif");
                filter.addExtension("png");
                filter.addExtension("jpg");
                filter.addExtension("jpeg");
                filter.setDescription("Images");

                dlg.setFileFilter(filter);
                int returnVal = dlg.showOpenDialog(owner);
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                        result = dlg.getSelectedFile().getPath();
                }

                return result;
        }

        /**
         * Select image files using file selection dialog
         * @param title dialog title
         * @param images full paths of selected files
         * @return boolean true if image selection succeeded
         * @see javax.swing#JFileChooser
         * */
        public static boolean selectImages(Component owner, String title, String[] images) {
                JFileChooser dlg = new JFileChooser();
                dlg.setDialogTitle(title);
                ExampleFileFilter filter = new ExampleFileFilter();
                filter.addExtension("bmp");
                // currently no Tiff support!
                //filter.addExtension("tif");
                filter.addExtension("gif");
                filter.addExtension("png");
                filter.addExtension("jpg");
                filter.addExtension("jpeg");
                filter.setDescription("Images");

                dlg.setFileFilter(filter);
                dlg.setMultiSelectionEnabled(true);
                int returnVal = dlg.showOpenDialog(owner);
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                        File[] files = dlg.getSelectedFiles();
                        for (int i = 0; i < images.length; ++i) {
                                if (i >= files.length)
                                        images[i] = null;
                                else
                                        images[i] = files[i].getPath();
                        }
                        return true;
                }

                return false;
        }

        /**
         * Convert Image into pixel matrix
         * @return byte[][] array with image pixel values
         * @see java.awt.image#PixelGrabber
         * */
        public static byte[] getImageData(Image img, int w, int h) {
                int[] pixels = new int[w * h];
                byte[] result = new byte[w * h];
                PixelGrabber pg = new PixelGrabber(img, 0, 0, w, h, pixels, 0, w);
                try {
                        pg.grabPixels();
                }
                catch (InterruptedException e) {
                        System.err.println("Interrupted waiting for pixels!");
                        return null;
                }
                if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
                        System.err.println("Image fetch aborted or errored");
                        return null;
                }

                for (int i = 0; i < result.length; ++i)
                        result[i] = (byte) pixels[i];
                return result;
        }

        /**
         * Create image from byte dump
         * @return Image new image created from given pixel values
         * @see java.awt.Toolkit#createImage
         * @see java.awt.image#MemoryImageSource
         * */
        public static Image createImage(byte[] data, int width, int height) {
                if (data == null) return null;
                int[] data2 = new int[width * height];
                int offset;
                int color;

                for (int y = 0; y < height; ++y) {
                        offset = y * width;
                        for (int x = 0; x < width; ++x) {
                                int b = data[y * width + x] & 0xff; // convert to unsigned
                                color = (0xff  << 8) | b;
                                color = (color << 8) | b;
                                color = (color << 8) | b;
                                // 0xffrrggbb - color format
                                // 0xff - color is opaque
                                data2[offset + x] = color;
                        }
                }

                return Toolkit.getDefaultToolkit().createImage(
                                new MemoryImageSource(
                                        width,
                                        height,
                                        data2,
                                        0,
                                        width));
        }
}
