unit VFFileList;

interface

uses SysUtils, Classes, VFinger, Utils, Math;

type
  PVFFileListItem = ^TVFFileListItem;
  TVFFileListItem = record
    Filename: string;
    Resolution: Integer;
    FingerId: string;
  end;

  TVFFileList = class
  private
    FItems: TList;
    function GetCount: Integer;
    function GetItem(Index: Integer): PVFFileListItem;
  public
    constructor Create;
    destructor Destroy; override;
    property Count: Integer read GetCount;
    property Item[Index: Integer]: PVFFileListItem read GetItem; default;
    procedure Clear;
    procedure Add; overload;
    procedure Add(Filename: string; Resolution: Integer = VF_IMAGE_RESOLUTION; FingerId: string = ''); overload;
    procedure Insert(Index: Integer); overload;
    procedure Insert(Index: Integer; Filename: string; Resolution: Integer = VF_IMAGE_RESOLUTION; FingerId: string = ''); overload;
    procedure Move(Index, NewIndex: Integer);
    procedure Remove(Index: Integer);
    function Load(Filename: string): Boolean; overload;
    procedure Load(Stream: TStream); overload;
    function Save(Filename: string): Boolean; overload;
    procedure Save(Stream: TStream); overload;
    class function Filter: string;
    class function FilterName: string;
    class function FilterString(FilterInName: Boolean = True): string;
    class function LoadFilterString(FilterInName: Boolean = True; AllFiles: Boolean = True): string;
    class function SaveFilterString(FilterInName: Boolean = True): string;
  end;

const
  DefaultFileFingerID = '$(NameNoExt)';
  DefaultScannerFingerID = '$(Name) $(Date) $(Time)';

function MakeFingerID(FingerId, Source: string; FromScanner: Boolean): string;
procedure CheckFingerID(FingerId: string; FromScanner: Boolean);
function GetFingerIDHelp(FromScanner: Boolean): string;

implementation

{ TVFFileList }

procedure TVFFileList.Add;
var
  Item: PVFFileListItem;
begin
  New(Item);
  Item.Resolution := VF_IMAGE_RESOLUTION;
  try
    FItems.Add(Item);
  except
    Dispose(Item);
    raise;
  end;
end;

procedure TVFFileList.Add(Filename: string; Resolution: Integer;
  FingerId: string);
var
  Item: PVFFileListItem;
begin
  New(Item);
  Item.Filename := Filename;
  Item.Resolution := Resolution;
  Item.FingerId := FingerId;
  try
    FItems.Add(Item);
  except
    Dispose(Item);
    raise;
  end;
end;

procedure TVFFileList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    Dispose(Item[I]);
  FItems.Clear;
end;

constructor TVFFileList.Create;
begin
  FItems := TList.Create;
end;

destructor TVFFileList.Destroy;
begin
  Clear;
  FItems.Free;
  inherited;
end;

class function TVFFileList.Filter: string;
begin
  Result := '*.vffl';
end;

class function TVFFileList.FilterName: string;
begin
  Result := 'File lists';
end;

class function TVFFileList.FilterString(FilterInName: Boolean): string;
begin
  Result := MakeFileFilterString(FilterName, Filter, FilterInName);
end;

function TVFFileList.GetCount: Integer;
begin
  Result := FItems.Count;
end;

function TVFFileList.GetItem(Index: Integer): PVFFileListItem;
begin
  Result := PVFFileListItem(FItems[Index]);
end;

procedure TVFFileList.Insert(Index: Integer);
var
  Item: PVFFileListItem;
begin
  New(Item);
  Item.Resolution := VF_IMAGE_RESOLUTION;
  try
    FItems.Insert(Index, Item);
  except
    Dispose(Item);
    raise;
  end;
end;

procedure TVFFileList.Insert(Index: Integer; Filename: string;
  Resolution: Integer; FingerId: string);
var
  Item: PVFFileListItem;
begin
  New(Item);
  Item.Filename := Filename;
  Item.Resolution := Resolution;
  Item.FingerId := FingerId;
  try
    FItems.Insert(Index, Item);
  except
    Dispose(Item);
    raise;
  end;
end;

function TVFFileList.Load(Filename: string): Boolean;
var
  Stream: TStream;
begin
  Stream := nil;
  try
    Stream := TFileStream.Create(Filename, fmOpenRead or fmShareDenyWrite);
    Load(Stream);
    Result := True;
  except
    Result := False;
  end;
  Stream.Free;
end;

procedure TVFFileList.Load(Stream: TStream);
var
  S: string;
  I, J: Integer;
  Filename: string;
  Resolution: Integer;
  FingerId: string;

  function ReadString(Stream: TStream; var S: string): Boolean;
  const
    MaxLineLength = 1000;
  var
    Buf: array[0..MaxLineLength] of AnsiChar;
    L: Integer;
  begin
    L := 0;
    while (L < MaxLineLength) and (Stream.Read(Buf[L], SizeOf(AnsiChar)) = SizeOf(AnsiChar)) do
    begin
      Inc(L);
      if Buf[L - 1] = #13 then
      begin
        if Stream.Read(Buf[L - 1], SizeOf(AnsiChar)) = SizeOf(AnsiChar) then
          if Buf[L - 1] <> #10 then Stream.Seek(-SizeOf(AnsiChar), soFromCurrent);
        Buf[L - 1] := #13;
        Break;
      end;
    end;
    Buf[L] := #0;
    Result := L <> 0;
    S := Buf;
  end;

begin
  Clear;
  while ReadString(Stream, S) do
  begin
    I := IndexOf(S, #13);
    if I <> -1 then SetLength(S, I - 1);
    if S <> EmptyStr then
    begin
      Resolution := VF_IMAGE_RESOLUTION;
      I := IndexOf(S, #9);
      if I <> -1 then
      begin
        J := IndexOf(S, #9, I + 1);
        if J <> -1 then
        begin
          FingerId := Copy(S, J + 1, Length(S) - J);
        end
        else J := Length(S) + 1;
        try
          Resolution := StrToInt(Copy(S, I + 1, J - I - 1));
        except
        end;
      end
      else I := Length(S) + 1;
      Filename := Copy(S, 1, I - 1);
      if Filename <> EmptyStr then Add(Filename, Resolution, FingerId);
    end;
  end;
end;

class function TVFFileList.LoadFilterString(FilterInName,
  AllFiles: Boolean): string;
begin
  AddFileFilterString(Result, FilterString(FilterInName));
  if AllFiles then AddAllFilesFilterString(Result, FilterInName);
end;

procedure TVFFileList.Move(Index, NewIndex: Integer);
var
  Item: PVFFileListItem;
begin
  Item := Self.Item[Index];
  FItems.Delete(Index);
  FItems.Insert(NewIndex, Item);
end;

procedure TVFFileList.Remove(Index: Integer);
begin
  FItems.Delete(Index);
end;

function TVFFileList.Save(Filename: string): Boolean;
var
  Stream: TStream;
begin
  Stream := nil;
  try
    Stream := TFileSTream.Create(Filename, fmCreate or fmShareDenyWrite);
    Save(Stream);
    Result := True;
  except
    Result := False;
  end;
  Stream.Free;
end;

procedure TVFFileList.Save(Stream: TStream);
var
  I, C: Integer;
  Item: PVFFileListItem;

  procedure WriteString(Stream: TStream; S: string);
  var
    L, I: Integer;
    Ch: AnsiChar;
  begin
    L := Length(S);
    for I := 1 to L do
    begin
      Ch := S[I];
      Stream.WriteBuffer(Ch, SizeOf(AnsiChar));
      if Ch = #13 then
      begin
        Ch := #10;
        Stream.WriteBuffer(Ch, SizeOf(AnsiChar));
      end;
    end;
  end;

begin
  C := Count - 1;
  for I := 0 to C do
  begin
    Item := Self.Item[I];
    WriteString(Stream, Format('%s'#9'%d'#9'%s', [Item.Filename, Item.Resolution, Item.FingerId]));
    if I <> C then WriteString(Stream, #13);
  end;
end;

class function TVFFileList.SaveFilterString(FilterInName: Boolean): string;
begin
  AddFileFilterString(Result, FilterString(FilterInName));
end;

{ Finger ID functions }

type
  TFingerIDProc = function(FromScanner: Boolean; Arguments: string; Filename: string): string;

  PFingerIDFunction = ^TFingerIDFunction;
  TFingerIDFunction = record
    Name: string;
    ArgumentCount: Integer;
    FromFile: Boolean;
    FromScanner: Boolean;
    Proc: TFingerIDProc;
    Help: string;
  end;

function FingerID_GetArgument(Arguments: string; I: Integer): string; forward;
function FingerID_TranslateFunction(Expression, Source: string; FromScanner: Boolean): string; forward;

function FingerID_Path(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := Filename;
end;

function FingerID_Drive(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := ExtractFileDrive(Filename);
  if (Length(Result) <> 0) and (Result[Length(Result)] = ':') then
    Result := Copy(Result, 1, Length(Result) - 1);
end;

function FingerID_Dir(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := ExtractFilePath(Filename);
end;

function FingerID_Name(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := ExtractFileName(Filename);
end;

function FingerID_Ext(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := ExtractFileExt(Filename);
  if (Length(Result) <> 0) and (Result[1] = '.') then Result := Copy(Result, 2, Length(Result) - 1);
end;

function FingerID_NameNoExt(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := ChangeFileExt(ExtractFileName(Filename), '');
end;

function FingerID_PathNoExt(FromScanner: Boolean; Arguments, Filename: string): string;
begin
  Result := ChangeFileExt(Filename, '');
end;

function FingerID_Subdir(FromScanner: Boolean; Arguments, Filename: string): string;
var
  I: Integer;
begin
  I := StrToInt(FingerID_GetArgument(Arguments, 0));
  Result := ExtractFileSubDir(Filename, I);
end;

function FingerID_ScannerFullName(FromScanner: Boolean; Arguments, ScannerId: string): string;
begin
  Result := ScannerId;
end;

function FingerID_ScannerName(FromScanner: Boolean; Arguments, ScannerId: string): string;
var
  I: Integer;
begin
  I := IndexOf(ScannerId, '\');
  if I = -1 then I := Length(ScannerId) + 1;
  Result := Copy(ScannerId, 1, I - 1);
end;

function FingerID_ScannerID(FromScanner: Boolean; Arguments, ScannerId: string): string;
var
  I: Integer;
begin
  I := IndexOf(ScannerId, '\');
  if I = -1 then Result := EmptyStr
  else Result := Copy(ScannerId, I + 1, Length(ScannerId) - I);
end;

function FingerID_Left(FromScanner: Boolean; Arguments, Str: string): string;
var
  C: Integer;
begin
  Result := FingerID_TranslateFunction(FingerID_GetArgument(Arguments, 0), Str, FromScanner);
  C := StrToInt(FingerID_GetArgument(Arguments, 1));
  if Abs(C) > Length(Result) then if C < 0 then C := -Length(Result) else C := Length(Result);
  if C >= 0 then Result := Copy(Result, 1, C) else Result := Copy(Result, -C + 1, Length(Result) + C);
end;

function FingerID_Right(FromScanner: Boolean; Arguments, Str: string): string;
var
  C: Integer;
begin
  Result := FingerID_TranslateFunction(FingerID_GetArgument(Arguments, 0), Str, FromScanner);
  C := StrToInt(FingerID_GetArgument(Arguments, 1));
  if Abs(C) > Length(Result) then if C < 0 then C := -Length(Result) else C := Length(Result);
  if C >= 0 then Result := Copy(Result, Length(Result) - C + 1, C) else Result := Copy(Result, 1, Length(Result) + C);
end;

function FingerID_Center(FromScanner: Boolean; Arguments, Str: string): string;
var
  I, C: Integer;
begin
  Result := FingerID_TranslateFunction(FingerID_GetArgument(Arguments, 0), Str, FromScanner);
  I := StrToInt(FingerID_GetArgument(Arguments, 1));
  if I = 0 then Exit;
  if I < 0 then
  begin
    I := Length(Result) + I;
    if I < 0 then I := 0;
  end
  else begin
    Dec(I);
    if I > Length(Result) then I := Length(Result);
  end;
  C := StrToInt(FingerID_GetArgument(Arguments, 2));
  if C >= 0 then Result := Copy(Result, I + 1, C) else Result := Copy(Result, 1, I)
    + Copy(Result, Min(I - C + 1, Length(Result)), Max(Length(Result) - I + C, 0));
end;

function FingerID_Date(FromScanner: Boolean; Arguments, Str: string): string;
begin
  Result := DateToStr(Now);
end;

function FingerID_Time(FromScanner: Boolean; Arguments, Str: string): string;
begin
  Result := TimeToStr(Now);
end;

const
  FingerIdMacroChar = '$';

  FingerIDFunctions: array[0..16] of TFingerIDFunction = (
    (Name: 'Path'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_Path;
      Help: 'Path - returns path to the file'),
    (Name: 'Drive'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_Drive;
      Help: 'Drive - returns drive letter from the path'),
    (Name: 'Dir'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_Dir;
      Help: 'Dir - returns file directory'),
    (Name: 'Name'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_Name;
      Help: 'Name - returns file name'),
    (Name: 'Ext'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_Ext;
      Help: 'Ext - returns file extension'),
    (Name: 'NameNoExt'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_NameNoExt;
      Help: 'NameNoExt - returns file name without extension'),
    (Name: 'PathNoExt'; ArgumentCount: 0; FromFile: True; FromScanner: False; Proc: FingerID_PathNoExt;
      Help: 'PathNoExt - returns file path without extension'),
    (Name: 'Subdir'; ArgumentCount: 1; FromFile: True; FromScanner: False; Proc: FingerID_Subdir;
      Help: 'SubDir(i) - if i = 0 returns all directories that the path contains'#13
      + ''#9'if i > 0 returns i directory from the left of the path'#13
      + ''#9'if i < 0 returns -i directory from the right of the path'),
    (Name: 'FullName'; ArgumentCount: 0; FromFile: False; FromScanner: True; Proc: FingerID_ScannerFullName;
      Help: 'FullName - returns full scanner name (with id)'),
    (Name: 'Name'; ArgumentCount: 0; FromFile: False; FromScanner: True; Proc: FingerID_ScannerName;
      Help: 'Name - returns scanner name'),
    (Name: 'ID'; ArgumentCount: 0; FromFile: False; FromScanner: True; Proc: FingerID_ScannerID;
      Help: 'ID - returns scanner id'),
    (Name: 'Left'; ArgumentCount: 2; FromFile: True; FromScanner: True; Proc: FingerID_Left;
      Help: 'Left(s, n) - if n > 0, returns at most n characters from the left of s'#13
      + ''#9'if n < 0, returns s without at most -n characters from the left'#13
      + ''#9'if n = 0, returns empty string'),
    (Name: 'Right'; ArgumentCount: 2; FromFile: True; FromScanner: True; Proc: FingerID_Right;
      Help: 'Right(s, n) - if n > 0, returns at most n characters from the right of s'#13
      + ''#9'if n < 0, returns s without at most -n characters from the right'#13
      + ''#9'if n = 0, returns empty string'),
    (Name: 'Center'; ArgumentCount: 3; FromFile: True; FromScanner: True; Proc: FingerID_Center;
      Help: 'Center(s, i, n) - if n > 0, returns at most n characters from the position'#13
      + ''#9'if n < 0, returns s without at most -n characters from the position'#13
      + ''#9'if n = 0, returns empty string'#13
      + ''#9'position is i from the left if i > 0 and -i from the right if i < 0'#13
      + ''#9'if i = 0 returns s'),
    (Name: 'Date'; ArgumentCount: 0; FromFile: False; FromScanner: True; Proc: FingerID_Date;
      Help: 'Date - returns current date'),
    (Name: 'Time'; ArgumentCount: 0; FromFile: False; FromScanner: True; Proc: FingerID_Time;
      Help: 'Time - returns current time'),
    (Name: ''; ArgumentCount: 1; FromFile: True; FromScanner: True; Proc: nil; Help: ''));

function FingerID_GetArguments(Expression: string; var I: Integer; Name: string; SpacesOk: Boolean = True): string;
var
  Str: string;
  S, E: Integer;
  C: Integer;
  J: Integer;
begin
  try
    S := 0; E := 0;
    C := 0;
    for J := I to Length(Expression) do
      if Expression[J] = '(' then
      begin
        if C = 0 then S := J;
        Inc(C);
      end
      else if Expression[J] = ')' then
      begin
        Dec(C);
        if C < 0 then Break;
        if C = 0 then
        begin
          E := J;
          Break;
        end;
      end
      else if C = 0 then
        if Expression[J] = ' ' then
        begin
          if not SpacesOk then raise Exception.Create('Unexpected space');
        end
        else raise Exception.CreateFmt('Unexpected char ''%c''', [Expression[J]]);
    if C > 0 then raise Exception.Create('Opening bracket without closing bracket');
    if C < 0 then raise Exception.Create('Closing bracket without opening bracket');
    if (S = 0) or (E = 0) then raise Exception.Create('Brackets not found');
    I := E + 1;
    Result := Copy(Expression, S + 1, E - S - 1);
    Result := Trim(Result);
    while (Length(Result) > 1) and (Result[1] = '(') and (Result[Length(Result)] = ')') do
    begin
      Result := Copy(Result, 2, Length(Result) - 2);
      Result := Trim(Result);
    end;
  except
    on E: Exception do
    begin
      if Name = EmptyStr then Str := 'macro' else Str := 'function ' + Name;
      raise Exception.Create(Format('%s after %s', [E.Message, Str]));
    end;
  end;
end;

function FingerID_ArgumentCount(Arguments: string): Integer;
var
  BC: Integer;
  J: Integer;
begin
  if Arguments = EmptyStr then Result := 0
  else begin
    BC := 0;
    Result := 0;
    for J := 1 to Length(Arguments) + 1 do
    begin
      if ((J = Length(Arguments) + 1) or (Arguments[J] = ',')) and (BC = 0) then Inc(Result)
      else if Arguments[J] = '(' then Inc(BC)
      else if Arguments[J] = ')' then Dec(BC);
    end;
  end;
end;

function FingerID_GetArgument(Arguments: string; I: Integer): string;
var
  C: Integer;
  II, J, S: Integer;
begin
  C := 0;
  II := I;
  S := 1;
  for J := 1 to Length(Arguments) + 1 do
  begin
    if ((J = Length(Arguments) + 1) or (Arguments[J] = ',')) and (C = 0) then
    begin
      if I = 0 then
      begin
        Result := Copy(Arguments, S, J - S);
        Result := Trim(Result);
        while (Length(Result) > 1) and (Result[1] = '(') and (Result[Length(Result)] = ')') do
        begin
          Result := Copy(Result, 2, Length(Result) - 2);
          Result := Trim(Result);
        end;
	Exit;
      end
      else begin
        Dec(I);
        S := J + 1;
      end;
    end
    else if Arguments[J] = '(' then Inc(C)
    else if Arguments[J] = ')' then Dec(C);
  end;
  raise Exception.CreateFmt('Argument %d not found', [ii]);
end;

procedure FingerID_TestArgumentCount(Arguments: string; C: Integer; Name: string);
var
  S: string;
begin
  if FingerID_ArgumentCount(Arguments) <> C then
  begin
    if Name = EmptyStr then S := 'Macro' else S := Name + ' function';
    raise Exception.CreateFmt('%s requires %s', [S, IntToString(C, 'argument', 'arguments', 'no')]);
  end;
end;

procedure FingerID_TestNoArguments(Expression: string; I: Integer; Name: string);
begin
  for I := I to Length(Expression) do
    if Expression[I] <> ' ' then raise Exception.CreateFmt('Unexpected char ''%c'' after function %s',
      [Expression[I], Name]);
end;

function FingerID_TranslateFunction(Expression, Source: string; FromScanner: Boolean): string;
var
  E, I: Integer;
  S, Arguments: string;
  F: PFingerIDFunction;
begin
  E := Length(Expression) + 1;
  for I := 1 to Length(Expression) do
    if (Expression[I] = '(') or (Expression[I] = ')') or (Expression[I] = ' ') then
    begin
      E := I;
      Break;
    end;
  S := Copy(Expression, 1, E - 1);
  if S = EmptyStr then raise Exception.Create('Function name expected');
  F := @FingerIDFunctions[0];
  while F.Name <> EmptyStr do
  begin
    if (FromScanner and F.FromScanner) or (not FromScanner and F.FromFile) then
      if CompareText(S, F.Name) = 0 then
      begin
        if F.ArgumentCount <> 0 then
        begin
          Arguments := FingerID_GetArguments(Expression, E, F.Name);
          FingerID_TestArgumentCount(Arguments, F.ArgumentCount, F.Name);
        end
        else begin
          FingerID_TestNoArguments(Expression, E, F.Name);
          Arguments := EmptyStr;
        end;
        Result := F.Proc(FromScanner, Arguments, Source);
        Exit;
      end;
    Inc(F);
  end;
  raise Exception.CreateFmt('Unknown function ''%s''', [S]);
end;

function MakeFingerID(FingerId, Source: string; FromScanner: Boolean): string;
var
  I, C: Integer;
  Arguments: string;
begin
  Result := EmptyStr;
  C := Length(FingerID);
  I := 1;
  while I <= C do
  begin
    if FingerId[I] = FingerIDMacroChar then
    begin
      Inc(I);
      if (I <> C + 1) and (FingerId[I] = FingerIDMacroChar) then
      begin
        Result := Result + FingerIDMacroChar;
      end
      else begin
        Arguments := FingerID_GetArguments(FingerId, I, EmptyStr, False);
        Dec(I);
        FingerID_TestArgumentCount(Arguments, 1, EmptyStr);
        Result := Result + FingerID_TranslateFunction(Arguments, Source, FromScanner);
      end;
    end
    else if FingerId[I] = '/' then
    begin
      if (I <> C - 2) and (FingerId[I + 1] = '/') then
      begin
        Inc(I);
        Result := Result + '/';
      end
      else if (Result = EmptyStr) or (Result[Length(Result)] <> '\') then Result := Result + '\';
    end
    else Result := Result + FingerId[I];
    Inc(I);
  end;
end;

procedure CheckFingerID(FingerId: string; FromScanner: Boolean);
begin
  MakeFingerID(FingerId, EmptyStr, FromScanner);
end;

function GetFingerIDHelp(FromScanner: Boolean): string;
var
  F: PFingerIDFunction;
const
  Intro: string =
    'When enrolling fingerprints, finger id is stored in the database along with features extracted from the fingerprint image.'#13 +
    'Fingerprints from different fingers must have different finger id. Same finger fingerprints must have the same finger id.'#13 +
    'When identifying a fingerprint in the database you will get identified finger id as a result.'#13 +
    'You can specify finger id directly for each fingerprint, like ''John''s thumb'' or use macros to make finger id from fingerprint image file name.'#13 +
    'Finger ID can contain any combination of characters and macros.'#13 +
    'Macros have to be in form of ''$(s)'', ''s'' can be one of the following functions:';
  Example: string =
    'Macros are especially usefull when enrolling a large amount of fingerprints from files'#13 +
    'For example, if you have fingerprint images in files that are named \"xxxxy.bmp\", where \"xxxx\" is finger number and \"y\" is scanning number,'#13 +
    'then you can specify \"$(Right(NameNoExt,-1))\" macro to get \"xxxx\" as finger id from the file name';
  Details: string =
    'If you want to have ''$'' character in finger id, use ''$$'' instead.'#13 +
    'If you specify ''/'' character, then ''\'' character is added to finger id if none is currently present.'#13 +
    'If you want to have the ''/'' character in finger id, use ''//'' instead.';
begin
  F := @FingerIDFunctions[0];
  Result := Intro;
  Result := Result + #13;
  while F.Name <> EmptyStr do
  begin
    if (FromScanner and F.FromScanner) or (not FromScanner and F.FromFile) then
    begin
      Result := Result + #13;
      Result := Result + F.Help;
    end;
    Inc(F);
  end;
  Result := Result + #13#13'i and n parameters of the functions have to be an integer values.';
  if not FromScanner then
  begin
    Result := Result + #13#13;
    Result := Result + Example;
  end
end;

end.

