﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Reflection;

namespace Neurotec.Biometrics.VeriFingerDemo
{
    public static class PathUtils
    {
        private static readonly char[] directorySeparatorChars = new char[] { Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar };
        private const string upOneLevel = "..";

        public static string GetPathWithoutExtension(string path)
        {
            return Path.ChangeExtension(path, null);
        }

        public static string GetSubdirectoryName(string path, int index)
        {
            string[] directories = path.Split(directorySeparatorChars);
            int startIndex = directories[0].Length != 0 && directories[0][directories[0].Length - 1] == Path.VolumeSeparatorChar ? 1 : 0;
            int count = Math.Max(directories.Length - startIndex - 1, 0);
            if(index == 0)
            {
                string[] newDirectories = new string[count];
                Array.Copy(directories, startIndex, newDirectories, 0, count);
                return string.Join(new string(Path.DirectorySeparatorChar, 1), newDirectories);
            }
            else
            {
                if(Math.Abs(index) > count) return string.Empty;
                return index > 0 ? directories[startIndex + index - 1] : directories[startIndex + count + index];
            }
        }

        public static string RemoveRoot(string path, int index)
        {
            if(path == null) throw new ArgumentNullException("path");
            if(index < -1) throw new ArgumentOutOfRangeException("index", index, "Index is less than -1");
            int i = 0;

			if (index != -1 && path.Length > 2 && Array.IndexOf(directorySeparatorChars, path[0]) != -1 && Array.IndexOf(directorySeparatorChars, path[1]) != -1)
				index += 3;

			while(index != -1)
            {
                int newI = path.IndexOfAny(directorySeparatorChars, i);
                if(newI == -1) break;
                i = newI + 1;
                index--;
            }
            return i == 0 ? path : path.Substring(i);
        }

        public static string GetRelativePath(string path)
        {
            return GetRelativePath(path, Directory.GetCurrentDirectory());
        }

        public static string GetRelativePath(string path, string root)
        {
            return GetRelativePath(path, root, false);
        }

        public static string GetRelativeDirectoryPath(string path, string root)
        {
            return GetRelativePath(path, root, true);
        }

        private static string GetRelativePath(string path, string root, bool isDirectory)
        {
            if(path == null) throw new ArgumentNullException("path");
            if(root == null) throw new ArgumentNullException("root");
            if(string.Compare(Path.GetPathRoot(root), Path.GetPathRoot(path), true) != 0) return path;
            string[] rootDirs = root.Split(directorySeparatorChars);
            string[] pathDirs = (isDirectory ? path : Path.GetDirectoryName(path)).Split(directorySeparatorChars);
            int i = 1;
            for(; i < rootDirs.Length && i < pathDirs.Length; i++)
                if(string.Compare(rootDirs[i], pathDirs[i], true) != 0) break;
            StringBuilder s = new StringBuilder();
            for(int j = i; j < rootDirs.Length; j++)
            {
                if(j != i) s.Append(Path.DirectorySeparatorChar);
                s.Append(upOneLevel);
            }
            for(int j = i; j < pathDirs.Length; j++)
            {
                if(j != i) s.Append(Path.DirectorySeparatorChar);
                s.Append(pathDirs[j]);
            }
            if(!isDirectory)
            {
                if(s.Length != 0) s.Append(Path.DirectorySeparatorChar);
                s.Append(Path.GetFileName(path));
            }
            return s.ToString();
        }

        public static string GetAbsolutePath(string path)
        {
            return GetAbsolutePath(path, Directory.GetCurrentDirectory());
        }

        public static string GetAbsolutePath(string path, string root)
        {
            if(path == null) throw new ArgumentNullException("path");
            if(root == null) throw new ArgumentNullException("root");
            if(Path.IsPathRooted(path)) return path;
            string[] rootDirs = root.Split(directorySeparatorChars);
            string[] pathDirs = path.Split(directorySeparatorChars);
            int c = rootDirs.Length;
            int o = 0;
            for(int i = 0; i < pathDirs.Length; i++)
                if(pathDirs[i] == upOneLevel)
                {
                    if(c > 1) c--;
                }
                else
                {
                    o = i;
                    break;
                }
            StringBuilder s = new StringBuilder();
            for(int i = 0; i < c; i++)
            {
                if(i != 0) s.Append(Path.DirectorySeparatorChar);
                s.Append(rootDirs[i]);
            }
            for(int i = o; i < pathDirs.Length - 1; i++)
            {
                if(i != 0 || c != 0) s.Append(Path.DirectorySeparatorChar);
                s.Append(pathDirs[i]);
            }
            if(pathDirs[pathDirs.Length - 1].Trim() != string.Empty)
            {
                if(s.Length != 0) s.Append(Path.DirectorySeparatorChar);
                s.Append(pathDirs[pathDirs.Length - 1]);
            }
            return s.ToString();
        }

        private static string GetAssemblyDataPath(string appDataPath, Assembly assembly)
        {
            AssemblyName assemblyName = assembly.GetName();
            AssemblyCompanyAttribute[] companyAttributes = (AssemblyCompanyAttribute[])assembly.GetCustomAttributes(typeof(AssemblyCompanyAttribute), true);
            AssemblyProductAttribute[] productAttributes = (AssemblyProductAttribute[])assembly.GetCustomAttributes(typeof(AssemblyProductAttribute), true);
            AssemblyVersionAttribute[] versionAttributes = (AssemblyVersionAttribute[])assembly.GetCustomAttributes(typeof(AssemblyVersionAttribute), true);
            string company = companyAttributes.Length == 0 ? string.Empty : companyAttributes[0].Company + Path.DirectorySeparatorChar;
            string product = (productAttributes.Length == 0 ? assemblyName.Name : productAttributes[0].Product) + Path.DirectorySeparatorChar;
            string version = versionAttributes.Length == 0 ? assemblyName.Version.ToString() : versionAttributes[0].Version;
            return Path.Combine(appDataPath, string.Concat(company, product, version));
        }

        public static string GetAppDataPath()
        {
            return GetAppDataPath(Assembly.GetCallingAssembly());
        }

        public static string GetAppDataPath(Assembly assembly)
        {
            return GetAssemblyDataPath(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), assembly);
        }

        public static string GetLocalAppDataPath()
        {
            return GetLocalAppDataPath(Assembly.GetCallingAssembly());
        }

        public static string GetLocalAppDataPath(Assembly assembly)
        {
            return GetAssemblyDataPath(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), assembly);
        }

        public static string GetCommonAppDataPath()
        {
            return GetCommonAppDataPath(Assembly.GetCallingAssembly());
        }

        public static string GetCommonAppDataPath(Assembly assembly)
        {
            return GetAssemblyDataPath(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), assembly);
        }

        public static void CreatePathToFile(string fileName)
        {
            Directory.CreateDirectory(Path.GetDirectoryName(fileName));
        }
    }
}