#include <stdlib.h>
#include <stdio.h>
#include <sys/timeb.h>

#if defined(__APPLE__)
#include <sys/time.h>
int ftime(struct timeb *tp)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	tp->time = tv.tv_sec;
	tp->millitm = tv.tv_usec / 1000;
	tp->timezone = 0;
	tp->dstflag = 0;

	return 0;
}
#endif

#include <VFExtractor.h>
#include <VFMatcher.h>
#include "img.h"

#define SAMPLES_PATH "../images/"
#define IMG_EXT "tif"

#ifdef __cplusplus
extern "C"
#endif

char *verifFileName[] = {
	SAMPLES_PATH "verif1."IMG_EXT,
	SAMPLES_PATH "verif2."IMG_EXT,
};

char *identFileName[] = {
	SAMPLES_PATH "ident1."IMG_EXT,
	SAMPLES_PATH "ident2."IMG_EXT,
	SAMPLES_PATH "ident3."IMG_EXT,
	SAMPLES_PATH "ident4."IMG_EXT,
	SAMPLES_PATH "ident5."IMG_EXT,
};

char *genFileName[] = {
	SAMPLES_PATH "gen1."IMG_EXT,
	SAMPLES_PATH "gen2."IMG_EXT,
	SAMPLES_PATH "gen3."IMG_EXT,
	SAMPLES_PATH "gent."IMG_EXT,
};

char identFileNameTest[] = SAMPLES_PATH "ident_test."IMG_EXT;

struct timeb T1, T2; /* structs for measuring time */

char *image1 = NULL;
char *image2 = NULL;

void *records[5];
NSizeType recordSizes[5];
NSizeType maxRecordSize = 0;

void do_extraction(HVFExtractor hExtractor, char *fileName, void *record, NSizeType *recordSize)
{
	char *image = NULL;
	int width = 0;
	int height = 0;
	NResult result;
	HNImage hImg;
	int milisecs = 0;

	result = load_img_file (fileName, (char**)&image, &width, &height, &hImg);
	if (result != 0)
	{
		printf ("Failed to load image (%s). Exiting. %d\n", fileName, result);
		exit (1);
	}

	ftime (&T1);
	result = VfeExtractFromImage(hExtractor, hImg, nfpUnknown, nfitLiveScanPlain, record, maxRecordSize, recordSize);
	ftime (&T2);
	NImageFree(hImg);
	milisecs = ((T2.time - T1.time) * 1000) + T2.millitm - T1.millitm;
	printf ("%s,\t %d sec, %d millisec\n",
		fileName, milisecs / 1000, milisecs % 1000 );

	if (result != N_OK)
	{
		printf("VfeExtract: An error occured (code %d). Exiting.\n", result);
		exit(1);
	}

	if (*recordSize == 0)
	{
		printf ("WARNING: Image is low quality, matching would fail.\n");
	}
}

void do_verify(HVFMatcher hMatcher, void *record1, NSizeType size1, void *record2, NSizeType size2)
{
	NResult result;
	NInt score = 0;

	result = VfmVerify(hMatcher, record1, size1, record2, size2, NULL, &score);

	if (result != N_OK)
	{
		printf("VfmVerify: An error occured. Exiting. %d\n", result);
		exit(1);
	}

	if (score > 0)
	{
		printf("Fingerprints matched. Similarity: %d\n", score);
	}
	else
	{
		printf("Fingerprints didn't match.\n");
	}
}

void do_identifynext(HVFMatcher hMatcher, void *record, NSizeType size, int *is, int i)
{
	NResult result;
	NInt score = 0;

	result = VfmIdentifyNext(hMatcher, record, size, NULL, &score);

	if (result != N_OK)
	{
		printf("VfmVerify: An error occured. Exiting. %d\n", result);
		exit(1);
	}

	if (score > 0)
	{
		*is = i;
		printf("Fingerprints matched. Similarity: %d\n", score);
	}
	else
	{
		printf("Fingerprints didn't match.\n");
	}
}

void do_generalize(HVFExtractor hExtractor, void *records[], NSizeType recordSizes[])
{
	NResult result;
	NInt baseTemplateIndex;

	result = VfeGeneralize(hExtractor, 3/*NFG_MIN_TEMPLATES*/, (const void **)records, recordSizes,
			records[4], maxRecordSize, &recordSizes[4], &baseTemplateIndex);

	if (result != N_OK)
	{
		printf("VfeGeneralize: An error occured. Exiting. %d\n", result);
		exit(1);
	}
}

void print_version(HVFExtractor extractor, HVFMatcher matcher)
{
	NUInt versionH, versionL;
	int string_size, max_string_size = 0;
	char *lib_name;

	string_size = VfeGetParameter(extractor, VFEP_NAME, NULL);
	if (string_size <= 0)
	{
		printf ("VfeGetParameter returns %d, exitting.\n", string_size);
		exit(1);
	}
	max_string_size = string_size;

	string_size = VfeGetParameter(extractor, VFEP_COPYRIGHT, NULL);
	if (string_size <= 0)
	{
		printf ("VfeGetParameter returns %d, exitting.\n", string_size);
		exit(1);
	}
	if (string_size > max_string_size)
	{
		max_string_size = string_size ;
	}
	string_size = VfmGetParameter(matcher, VFMP_NAME, NULL);
	if (string_size <= 0)
	{
		printf ("VfmGetParameter returns %d, exitting.\n", string_size);
		exit(1);
	}
	if (string_size > max_string_size)
	{
		max_string_size = string_size;
	}

	string_size = VfmGetParameter(matcher, VFMP_COPYRIGHT, NULL);
	if (string_size <= 0)
	{
		printf ("VfmGetParameter returns %d, exitting.\n", string_size);
		exit(1);
	}
	if (string_size > max_string_size)
	{
		max_string_size = string_size ;
	}

	lib_name = malloc(max_string_size+1);
	if (lib_name == NULL)
	{
		printf ("out of memory.\n");
		exit(1);
	}

	VfeGetParameter(extractor, VFEP_NAME, lib_name);
	VfeGetParameter(extractor, VFEP_VERSION_HIGH, &versionH);
	VfeGetParameter(extractor, VFEP_VERSION_LOW, &versionL);
	printf("\n%s (version: %u.%u.%u.%u)\n", lib_name,
		versionH >> 16, versionH & 0xFFFF, versionL >> 16, versionL & 0xFFFF);
	VfeGetParameter(extractor, VFEP_COPYRIGHT, lib_name);
	printf("%s\n", lib_name);

	VfmGetParameter(matcher, VFMP_NAME, lib_name);
	VfmGetParameter(matcher, VFMP_VERSION_HIGH, &versionH);
	VfmGetParameter(matcher, VFMP_VERSION_LOW, &versionL);
	printf("%s (version: %u.%u.%u.%u)\n", lib_name,
		versionH >> 16, versionH & 0xFFFF, versionL >> 16, versionL & 0xFFFF);
	VfmGetParameter(matcher, VFMP_COPYRIGHT, lib_name);
	printf("%s\n", lib_name);
	printf("Algorithm demo.\n\n");
	free(lib_name);
}

int main (int argc, char *argv[])
{
	HVFExtractor	extractor;
	HVFMatcher	matcher;
	NResult		result;
	int		i, is = -1;

	result = VfeCreate(&extractor);
	if (result != N_OK)
	{
		printf ("VfeCreate returns %d\n", result);
		return result;
	}

	result = VfmCreate(&matcher);
	if (result != N_OK)
	{
		printf ("VfmCreate returns %d\n", result);
		return result;
	}

	print_version(extractor, matcher);

	printf ("\t\t\t**** Verification Demo ****\n"
		"> Will process two fingerprint images and see if they match. <\n"
		"Image file, image processing time\n");

	VfeGetMaxTemplateSize(extractor, &maxRecordSize);
	for (i = 0; i < 6; i++)
	{
		records[i] = malloc(maxRecordSize);
		if (records[i] == NULL)
		{
			printf("out of memory\n");
			return 1;
		}
	}

	do_extraction(extractor, verifFileName[0], records[0], &recordSizes[0]);
	do_extraction(extractor, verifFileName[1], records[1], &recordSizes[1]);
	do_verify(matcher, records[0], recordSizes[0], records[1], recordSizes[1]);

	printf ("\n\t\t\t**** Identification Demo ****\n"
		"> Will process five images and see which one of them matches the sixth. <\n"
		"Image file, image processing time\n");

	do_extraction(extractor, identFileNameTest, records[0], &recordSizes[0]);
	VfmIdentifyStart(matcher, records[0], recordSizes[0], NULL);

	for (i = 0; i < 5; i++)
	{
		do_extraction(extractor, identFileName[i], records[1], &recordSizes[1]);
		do_identifynext(matcher, records[1], recordSizes[1], &is, i);
	}
	VfmIdentifyEnd(matcher);

	if (is != -1)
		printf("The test fingerprint matched with %s.\n", identFileName[is]);
	else
		printf("The test fingerprint didn't match with any of the others.\n");

	printf ("\n\t\t\t**** Generalization Demo ****\n"
		"> Will generalize 3 images <\n"
		"Image file, image processing time\n");
	for (i = 0; i < 4; i++)
		do_extraction(extractor, genFileName[i], records[i], &recordSizes[i]);
	do_generalize(extractor, records, recordSizes);
	do_verify(matcher, records[4], recordSizes[4], records[3], recordSizes[3]);

	VfeFree(extractor);
	VfmFree(matcher);

	return 0;
}

